module hash.base;

class HashException : Exception
{
	this(char[] msg) { super(msg); }
}

class Digest
{
	abstract char[] toString();
}

class Hash
{
	this()
	{
		buffer = new ubyte[blockSize()];
	}
	
	Digest sum(void[] data)
	{
		start();
		update(data);
		return finish();
	}

	void start()
	{
		bytes = 0;
	}
	
	void update(void[] input)
	{
		ubyte[] data = cast(ubyte[])input;
		uint i;
		
		i = bytes & (blockSize-1);
		bytes += data.length;

		if (data.length+i < blockSize) {
			buffer[i..i+data.length] = data[];
			return ;
		}
		
		buffer[i..blockSize] = data[0..blockSize-i];
		transform(buffer);

		for(i = blockSize-i; i+blockSize-1 < data.length; i += blockSize)
			transform(data[i..i+blockSize]);

		buffer[0..data.length-i] = data[i..data.length];
	}
	
	Digest finish()
	{
		ulong i;
	
		i = bytes & (blockSize-1);
		if (i < blockSize-addSize) padMessage(buffer[i..blockSize-addSize]);
		else {
			padMessage(buffer[i..blockSize]);
			transform(buffer);
			(cast(ubyte[])buffer)[] = 0;
		}
		
		padLength(buffer[blockSize-addSize..blockSize],bytes);
		transform(buffer);

		extend();
	
		return getDigest();
	}
	
	abstract Digest getDigest();
	
protected:
	abstract uint blockSize();
	abstract uint addSize();
	abstract void padMessage(ubyte[] data);
	abstract void padLength(ubyte[] data, ulong length) {}
	abstract void transform(ubyte[] data);
	abstract void extend() {}	
	
private:
	ubyte[] buffer;
	uint bytes;
}

void littleEndian32(ubyte[] input, uint[] output)
{
	output[] = 0;

	assert(output.length == input.length/4);
	for(uint i = 0; i < input.length; i++) {
		output[i/4] <<= 8;
		version (BigEndian) output[i/4] |= cast(uint)input[i];
		else output[i/4] |= cast(uint)input[i^3];
	}
}

void bigEndian32(ubyte[] input, uint[] output)
{		
	output[] = 0;

	assert(output.length == input.length/4);
	for(uint i = 0; i < input.length; i++) {
		output[i/4] <<= 8;
		version (BigEndian) output[i/4] |= cast(uint)input[i^3];
		else output[i/4] |= cast(uint)input[i];
	}
}

void littleEndian64(ubyte[] input, ulong[] output)
{
	output[] = 0;

	assert(output.length == input.length/8);
	for(uint i = 0; i < input.length; i++) {
		output[i/8] <<= 8;
		version (BigEndian) output[i/8] |= cast(ulong)input[i];
		else output[i/8] |= cast(ulong)input[i^7];
	}
}


void bigEndian64(ubyte[] input, ulong[] output)
{
	output[] = 0;

	assert(output.length == input.length/8);
	for(uint i = 0; i < input.length; i++) {
		output[i/8] <<= 8;
		version (BigEndian) output[i/8] |= cast(ulong)input[i^7];
		else output[i/8] |= cast(ulong)input[i];
	}
}


uint rotateLeft(uint x, uint n)
{
	version (X86) {
		asm {
			naked           ;
			mov	ECX,EAX     ;
			mov	EAX,4[ESP]  ;
			rol	EAX,CL      ;
			ret	4		    ;
		}
	}
	else {
		return (x << n) | (x >> (32-n));
	}
}

private import std.string;

template toHexString(T){ char[] toHexString(T d) {
	char[] result;

	if (d.length != 0) {
		typeof(d[0]) u;
		uint sz = u.sizeof*2;
		uint ndigits = 0;

		result = new char[sz*d.length];
		for(int i = d.length-1; i >= 0; i--) {			
			u = d[i];
			for(; u; u /= 16) {
				result[result.length-1-ndigits] = hexdigits[u & 15];
				ndigits++;
			}
			for(; ndigits < (d.length-i)*sz; ndigits++)
				result[result.length-1-ndigits] = '0';
		}
	}

	return result;
}}
